/**
 * An LCD time display/Controller based on I2C (TWI)
 * Author: Kerry Wong
 * Website: http://www.kerrywong.com
 *
 * The clock is synchronized with the remote host where an RTC is used to
 * keep track of time. The time/alarm can be set via the buttons and they
 * can be synced back to the host via I2C.
 */

#define __AVR_ATmega328P__

/** <editor-fold desc="Includes"> */
#include <binary.h>
#include <HardwareSerial.h>
#include <pins_arduino.h>
#include <WConstants.h>
#include <wiring.h>
#include <wiring_private.h>
#include <WProgram.h>
#include <EEPROM/EEPROM.h>
#include <LiquidCrystal/LiquidCrystal.h>
#include <Wire/Wire.h>
#include <Wire/utility/twi.h>
#include <avr/interrupt.h>
/** </editor-fold> */

/** <editor-fold desc="Pin Definitions"> */
const int PIN_BACKLIGHT = 6;
const int PIN_LCD_RS = 12;
const int PIN_LCD_EN = 11;
const int PIN_LCD_D4 = 5;
const int PIN_LCD_D5 = 4;
const int PIN_LCD_D6 = 3;
const int PIN_LCD_D7 = 2;

const int PIN_ALARM = 10; //chip pin 16
const int PIN_RELAY_1 = 14; //chip pin 23
const int PIN_RELAY_2 = 15; //chip pin 24
const int PIN_RELAY_3 = 16; //chip pin 25
const int PIN_RELAY_4 = 17; //chip pin 26
const int PIN_RELAY_5 = 13; //chip pin 19

const int NUM_OF_ALARMS = 1;
const int NUM_OF_RELAYS = 5;
int alarmPins[] = {PIN_ALARM, PIN_RELAY_1, PIN_RELAY_2, PIN_RELAY_3, PIN_RELAY_4, PIN_RELAY_5};
/** </editor-fold> */

/** <editor-fold desc="Button Definitions"> */
const int NUM_OF_BUTTONS = 5;

const int BTN_MODE_SELECT = 0;
const int BTN_CURSOR_CTRL = 1;
const int BTN_DOWN = 2;
const int BTN_UP = 3;
const int BTN_ENTER = 4;

const int BTN_PINS[] = {0, 1, 7, 8, 9};

boolean btnPressed[] = {false, false, false, false, false};
int btnState[] = {0, 0, 0, 0, 0};
/** </editor-fold> */

/** <editor-fold desc="Setup Modes"> */
const int MODE_NORMAL_CANCEL = 0;
const int MODE_TIME_SETUP = 1;
const int MODE_ALARM_SETUP = 2;
const int MODE_RELAY_SETUP = 3;

int currentMode = MODE_NORMAL_CANCEL;
/** </editor-fold> */

/** <editor-fold desc="Cursor Positions"> */
const int CURSOR_TIME_MONTH = 0;
const int CURSOR_TIME_DAY = 3;
const int CURSOR_TIME_DOW = 6;
const int CURSOR_TIME_HOUR = 10;
const int CURSOR_TIME_MIN = 13;
int timeCursor = CURSOR_TIME_MIN;

const int CURSOR_ALARM_ON_HOUR = 0;
const int CURSOR_ALARM_ON_MIN = 3;
const int CURSOR_ALARM_OFF_HOUR = 6;
const int CURSOR_ALARM_OFF_MIN = 9;
const int CURSOR_ALARM_ON_OFF_IND = 12;
int alarmCursor = CURSOR_ALARM_ON_HOUR;
/** </editor-fold> */

/** <editor-fold desc="TWI Addresses"> */
const int ADDR_TWI_LOCAL = 4;
const int ADDR_TWI_REMOTE = 5;
/** </editor-fold> */

/** <editor-fold desc="Booleans"> */
volatile boolean intProc = false;

volatile boolean blinkMinute = false;
volatile boolean blinkHour = false;
volatile boolean blinkDOW = false;
volatile boolean blinkDay = false;
volatile boolean blinkMonth = false;

volatile boolean blinkAlarmOnMinute = false;
volatile boolean blinkAlarmOnHour = false;
volatile boolean blinkAlarmOffMinute = false;
volatile boolean blinkAlarmOffHour = false;
volatile boolean blinkAlarmOnOffInd = false;

boolean backLight = false;
/** </editor-fold> */

/** <editor-fold desc="LCD Symbols"> */
/* defines the symbol for clock setup */
byte CLOCK_SYM = 0;
byte SYMBOL_CLOCK[8] = {
    B00000,
    B00100,
    B00100,
    B00100,
    B00100,
    B01000,
    B10000,
};

/* defines the symbol for clock setup (selected) */
byte CLOCK_SYM_SEL = 1;
byte SYMBOL_CLOCK_SEL[8] = {
    B11111,
    B00100,
    B00100,
    B00100,
    B00100,
    B01000,
    B10000,
};

/* defines the symbol for alarm setup */
byte ALARM_SYM = 2;
byte SYMBOL_ALARM[8] = {
    B00000,
    B00000,
    B01000,
    B10010,
    B10101,
    B10010,
    B01000,
};

/* defines the symbol for alarm setup (selected) */
byte ALARM_SYM_SEL = 3;
byte SYMBOL_ALARM_SEL[8] = {
    B11111,
    B00000,
    B01000,
    B10010,
    B10101,
    B10010,
    B01000,
};

/* defines the symbol for relay control setup */
byte CTRL_SYM = 4;
byte SYMBOL_CTRL[8] = {
    B00000,
    B00000,
    B00010,
    B00100,
    B01110,
    B00100,
    B01000,
};

/* defines the symbol for relay control setup (selected)*/
byte CTRL_SYM_SEL = 5;
byte SYMBOL_CTRL_SEL[8] = {
    B11111,
    B00000,
    B00010,
    B00100,
    B01110,
    B00100,
    B01000,
};

/* defines the symbol to indicate a setting has been saved */
byte SAVED_SYM = 6;
byte SYMBOL_SAVED[8] = {
    B01110,
    B01010,
    B01010,
    B11111,
    B10001,
    B10101,
    B11111,
};

/* defines where the symbol starts on the second line */
const int SYM_LINE_OFFSET = 4;
/** </editor-fold> */

/** <editor-fold desc="Enums"> */
enum RTCMessageType {
    GetTime,
    SetTime
};
/** </editor-fold> */

/** <editor-fold desc="Structs"> */
/* defines the message passed between the two MCUs */
struct RTCMessage {
    byte sender;
    byte msgType;
    byte year;
    byte month;
    byte day;
    byte hour;
    byte minute;
    byte secondInd;
    byte dayOfWeek;
};

struct AlarmData {
    byte alarmNum;
    byte onHour;
    byte onMinute;
    byte offHour;
    byte offMinute;
    byte alarmSet;
};
/** </editor-fold> */

/** <editor-fold desc="Other Global Variables"> */
RTCMessage message, messageCopy;
AlarmData alarms[8];

char DOW[8][4] = {"   ", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

char defaultDisplayMsg[] = "--/-- --- --:--";

LiquidCrystal lcd(PIN_LCD_RS, PIN_LCD_EN, PIN_LCD_D4, PIN_LCD_D5, PIN_LCD_D6, PIN_LCD_D7);
/** </editor-fold> */

/** same as delay() but works with interrupts */
void delayMilliseconds(int ms) {
    for (int i = 0; i < ms; i++) {
        delayMicroseconds(1000);
    }
}

/** detects whether the ith button is pressed */
boolean buttonPressed(int btnIdx) {
    if (digitalRead(BTN_PINS[btnIdx]) == LOW && !btnPressed[btnIdx]) {
        delayMilliseconds(20);
        if (digitalRead(BTN_PINS[btnIdx]) == LOW) {
            btnPressed[btnIdx] = true;
            return true;
        }
    }

    return false;
}

/** detects whether the ith button is released */
boolean buttonRleased(int btnIdx) {
    if (digitalRead(BTN_PINS[btnIdx]) == HIGH && btnPressed[btnIdx]) {
        delayMilliseconds(20);
        if (digitalRead(BTN_PINS[btnIdx]) == HIGH) {
            btnPressed[btnIdx] = false;
            return true;
        }
    }
    return false;
}

/** displays time */
void displayTime() {
    lcd.setCursor(0, 0);

    char s[20];
    char formatStr1[] = "%2d/%02d %s %2d:%02d";
    char formatStr2[] = "%2d/%02d %s %2d %02d";

    char formatStrMinuteEdt[] = "%2d/%02d %s %2d:  ";
    char formatStrHourEdt[] = "%2d/%02d %s   :%02d";
    char formatStrDOWEdt[] = "%2d/%02d     %2d:%02d";
    char formatStrDayEdt[] = "%2d/   %s %2d:%02d";
    char formatStrMonthEdt[] = "  /%02d %s %2d:%02d";

    if (currentMode == MODE_NORMAL_CANCEL) {
        messageCopy = message;
    }

    /* depending on whether it is in edited mode, Hour/Minute/DOW/Day/Month blinks */
    if (message.secondInd == 1) {
        sprintf(s, formatStr1, messageCopy.month, messageCopy.day, DOW[messageCopy.dayOfWeek], messageCopy.hour, messageCopy.minute);
    } else {
        if (blinkMinute)
            sprintf(s, formatStrMinuteEdt, messageCopy.month, messageCopy.day, DOW[messageCopy.dayOfWeek], messageCopy.hour);
        else if (blinkHour)
            sprintf(s, formatStrHourEdt, messageCopy.month, messageCopy.day, DOW[messageCopy.dayOfWeek], messageCopy.minute);
        else if (blinkDOW)
            sprintf(s, formatStrDOWEdt, messageCopy.month, messageCopy.day, messageCopy.hour, messageCopy.minute);
        else if (blinkDay)
            sprintf(s, formatStrDayEdt, messageCopy.month, DOW[messageCopy.dayOfWeek], messageCopy.hour, messageCopy.minute);
        else if (blinkMonth)
            sprintf(s, formatStrMonthEdt, messageCopy.day, DOW[messageCopy.dayOfWeek], messageCopy.hour, messageCopy.minute);
        else
            sprintf(s, formatStr2, messageCopy.month, messageCopy.day, DOW[messageCopy.dayOfWeek], messageCopy.hour, messageCopy.minute);
    }

    lcd.print(s);
}

/**
 * Displays alarm settings in the following format
 * HH:MM_HH:MM_ON/OFF */
void displayAlarm(int alarmNum) {
    lcd.setCursor(0, 0);

    char onHour[3] = "--", onMinute[3] = "--", offHour[3] = "--", offMinute[3] = "--", onOffInd[4] = "Off";
    char s[20];

    char formatStr[] = "%s:%s %s:%s %s";

    char formatStrOnHourEdt[] = "  :%s %s:%s %s";
    char formatStrOnMinuteEdt[] = "%s:   %s:%s %s";
    char formatStrOffHourEdt[] = "%s:%s   :%s %s";
    char formatStrOffMinuteEdt[] = "%s:%s %s:   %s";
    char formatStrOnOffIndEdt[] = "%s:%s %s:%s    ";

    if (alarms[alarmNum].onHour != 255) sprintf(onHour, "%2d", alarms[alarmNum].onHour);
    else sprintf(onHour, "%s", "--");
    if (alarms[alarmNum].onMinute != 255) sprintf(onMinute, "%02d", alarms[alarmNum].onMinute);
    else sprintf(onMinute, "%s", "--");
    if (alarms[alarmNum].offHour != 255) sprintf(offHour, "%2d", alarms[alarmNum].offHour);
    else sprintf(offHour, "%s", "--");
    if (alarms[alarmNum].offMinute != 255) sprintf(offMinute, "%02d", alarms[alarmNum].offMinute);
    else sprintf(offMinute, "%s", "--");
    if (alarms[alarmNum].alarmSet != 1) sprintf(onOffInd, "%s", "Off");
    else sprintf(onOffInd, "%s", "On ");

    sprintf(s, formatStr, onHour, onMinute, offHour, offMinute, onOffInd);

    if (message.secondInd == 1) {
        if (blinkAlarmOnHour) {
            sprintf(s, formatStrOnHourEdt, onMinute, offHour, offMinute, onOffInd);
        } else if (blinkAlarmOnMinute) {
            sprintf(s, formatStrOnMinuteEdt, onHour, offHour, offMinute, onOffInd);
        } else if (blinkAlarmOffHour) {
            sprintf(s, formatStrOffHourEdt, onHour, onMinute, offMinute, onOffInd);
        } else if (blinkAlarmOffMinute) {
            sprintf(s, formatStrOffMinuteEdt, onHour, onMinute, offHour, onOffInd);
        } else if (blinkAlarmOnOffInd) {
            sprintf(s, formatStrOnOffIndEdt, onHour, onMinute, offHour, offMinute);
        }
    }

    lcd.print(s);
}

/** receive an i2c event from the remote MCU */
void receiveEvent(int howMany) {
    intProc = true;

    message.sender = Wire.receive();
    message.msgType = Wire.receive();
    message.year = Wire.receive();
    message.month = Wire.receive();
    message.day = Wire.receive();
    message.hour = Wire.receive();
    message.minute = Wire.receive();
    message.secondInd = Wire.receive();
    message.dayOfWeek = Wire.receive();

    if (currentMode == MODE_TIME_SETUP || currentMode == MODE_NORMAL_CANCEL) {
        displayTime();
    } else if (currentMode >= MODE_ALARM_SETUP) {
        displayAlarm(currentMode - MODE_ALARM_SETUP);
    }

    intProc = false;
}

/* initializes the alarm EEPROM area
 * there are 5 alarms, and each alarm has 6 bytes
 * the first byte is the ID of the alarm
 * and the last byte indicate whether the alarm
 * is enabled. By default the last byte is 0 which
 * set the alarm to off mode.*/
void initAlarm(int n) {
    EEPROM.write(n * 6, n);
    EEPROM.write(n * 6 + 5, 0);
}

/* get all alarm data from memory */
void getAlarms() {
    for (int i = 0; i < NUM_OF_ALARMS + NUM_OF_RELAYS; i++) {
        int addr = i * 6;

        if (EEPROM.read(addr) == 255) {
            initAlarm(i);
        }

        alarms[i].alarmNum = EEPROM.read(addr);
        alarms[i].onHour = EEPROM.read(addr + 1);
        alarms[i].onMinute = EEPROM.read(addr + 2);
        alarms[i].offHour = EEPROM.read(addr + 3);
        alarms[i].offMinute = EEPROM.read(addr + 4);
        alarms[i].alarmSet = EEPROM.read(addr + 5);
    }
}

/**
 * Possible settings:
 * HH:MM HH:MM (onTime offTime, if offTime is less than
 *      onTime, the offTime takes effect the next day)
 * --:MM --:MM (Hourly alarm, if the offTime minute is less than
 *      onTime minute, the offTime takes effect the next hour)
 * If the alarm is set to be on, but the time formats are not one of the above,
 * the alarm is ignored.
 */
void checkAlarms() {
    for (int i = 0; i < NUM_OF_ALARMS + NUM_OF_RELAYS; i++) {
        if (alarms[i].alarmSet == 1) {// if alarm is set
            //Check time format
            if ((alarms[i].onHour != 255 && alarms[i].onMinute != 255 &&
                    alarms[i].offHour != 255 && alarms[i].offMinute != 255) ||
                    (alarms[i].onHour == 255 && alarms[i].onMinute != 255 &&
                    alarms[i].offHour == 255 && alarms[i].offMinute == 255)) {

                //check on condition
                if ((alarms[i].onHour == 255 || alarms[i].onHour == message.hour) &&
                        (message.minute >= alarms[i].onMinute)) {
                    digitalWrite(alarmPins[i], HIGH);
                }

                //check off condition
                if ((alarms[i].onHour == 255 || alarms[i].onHour == message.hour) &&
                        message.minute >= alarms[i].offMinute && alarms[i].onMinute <= alarms[i].offMinute) {
                    digitalWrite(alarmPins[i], LOW);
                }

                if ((alarms[i].onHour == 255 || alarms[i].onHour + 1 == message.hour ||
                        (alarms[i].onHour == 23 && message.hour == 0) &&
                        message.minute >= alarms[i].offMinute && alarms[i].onMinute > alarms[i].offMinute)) {
                    digitalWrite(alarmPins[i], LOW);
                }
            }
        }
    }
}

/**
 * renders the current display mode
 * depend on which mode it is actually in
 */
void renderCurrentMode() {
    if (currentMode > NUM_OF_ALARMS + NUM_OF_RELAYS + 1) {
        currentMode = MODE_NORMAL_CANCEL;
    }

    lcd.setCursor(SYM_LINE_OFFSET, 1);

    blinkMinute = false;
    blinkHour = false;
    blinkDOW = false;
    blinkDay = false;
    blinkMonth = false;

    lcd.write(CLOCK_SYM);

    for (int i = 0; i < NUM_OF_ALARMS; i++)
        lcd.write(ALARM_SYM);

    for (int i = 0; i < NUM_OF_RELAYS; i++)
        lcd.write(CTRL_SYM);

    if (currentMode == MODE_NORMAL_CANCEL) {

    } else if (currentMode == MODE_TIME_SETUP) {
        timeCursor = CURSOR_TIME_MIN;
        blinkMinute = true;
        lcd.setCursor(SYM_LINE_OFFSET, 1);
        lcd.write(CLOCK_SYM_SEL);
    } else if (currentMode >= MODE_ALARM_SETUP && currentMode < MODE_ALARM_SETUP + NUM_OF_ALARMS) {
        alarmCursor = CURSOR_ALARM_ON_HOUR;
        blinkAlarmOnHour = true;
        lcd.setCursor(SYM_LINE_OFFSET + currentMode - 1, 1);
        lcd.write(ALARM_SYM_SEL);
    } else if (currentMode >= MODE_RELAY_SETUP && currentMode < MODE_RELAY_SETUP + NUM_OF_RELAYS) {
        alarmCursor = CURSOR_ALARM_ON_HOUR;
        blinkAlarmOnHour = true;
        lcd.setCursor(SYM_LINE_OFFSET + currentMode - 1, 1);
        lcd.write(CTRL_SYM_SEL);
    }

}

/** renders time setup mode */
void renderTimeSetup() {
    switch (timeCursor) {
        case CURSOR_TIME_MIN:
            timeCursor = CURSOR_TIME_HOUR;
            blinkMinute = false;
            blinkHour = true;
            blinkDOW = false;
            blinkDay = false;
            blinkMonth = false;
            break;
        case CURSOR_TIME_HOUR:
            timeCursor = CURSOR_TIME_DOW;
            blinkMinute = false;
            blinkHour = false;
            blinkDOW = true;
            blinkDay = false;
            blinkMonth = false;
            break;
        case CURSOR_TIME_DOW:
            timeCursor = CURSOR_TIME_DAY;
            blinkMinute = false;
            blinkHour = false;
            blinkDOW = false;
            blinkDay = true;
            blinkMonth = false;
            break;
        case CURSOR_TIME_DAY:
            timeCursor = CURSOR_TIME_MONTH;
            blinkMinute = false;
            blinkHour = false;
            blinkDOW = false;
            blinkDay = false;
            blinkMonth = true;
            break;
        case CURSOR_TIME_MONTH:
            timeCursor = CURSOR_TIME_MIN;
            blinkMinute = true;
            blinkHour = false;
            blinkDOW = false;
            blinkDay = false;
            blinkMonth = false;
            break;
        default:
            break;
    }
}

/** renders alarm setup mode */
void renderAlarmSetup() {
    switch (alarmCursor) {
        case CURSOR_ALARM_ON_HOUR:
            alarmCursor = CURSOR_ALARM_ON_MIN;
            blinkAlarmOnHour = false;
            blinkAlarmOnMinute = true;
            blinkAlarmOffHour = false;
            blinkAlarmOffMinute = false;
            blinkAlarmOnOffInd = false;
            break;
        case CURSOR_ALARM_ON_MIN:
            alarmCursor = CURSOR_ALARM_OFF_HOUR;
            blinkAlarmOnHour = false;
            blinkAlarmOnMinute = false;
            blinkAlarmOffHour = true;
            blinkAlarmOffMinute = false;
            blinkAlarmOnOffInd = false;
            break;
        case CURSOR_ALARM_OFF_HOUR:
            alarmCursor = CURSOR_ALARM_OFF_MIN;
            blinkAlarmOnHour = false;
            blinkAlarmOnMinute = false;
            blinkAlarmOffHour = false;
            blinkAlarmOffMinute = true;
            blinkAlarmOnOffInd = false;
            break;
        case CURSOR_ALARM_OFF_MIN:
            alarmCursor = CURSOR_ALARM_ON_OFF_IND;
            blinkAlarmOnHour = false;
            blinkAlarmOnMinute = false;
            blinkAlarmOffHour = false;
            blinkAlarmOffMinute = false;
            blinkAlarmOnOffInd = true;
            break;
        case CURSOR_ALARM_ON_OFF_IND:
            alarmCursor = CURSOR_ALARM_ON_HOUR;
            blinkAlarmOnHour = true;
            blinkAlarmOnMinute = false;
            blinkAlarmOffHour = false;
            blinkAlarmOffMinute = false;
            blinkAlarmOnOffInd = false;
            break;
    }
}

/* adjust time setup */
void adjustTimeSetup(int btn) {
    if (timeCursor == CURSOR_TIME_MIN) {
        if (btn == BTN_UP) {
            if (messageCopy.minute >= 59) messageCopy.minute = 0;
            else messageCopy.minute++;
        } else {
            if (messageCopy.minute <= 0) messageCopy.minute = 59;
            else messageCopy.minute--;
        }
    } else if (timeCursor == CURSOR_TIME_HOUR) {
        if (btn == BTN_UP) {
            if (messageCopy.hour >= 23) messageCopy.hour = 0;
            else messageCopy.hour++;
        } else {
            if (messageCopy.hour <= 0) messageCopy.hour = 23;
            else messageCopy.hour--;
        }
    } else if (timeCursor == CURSOR_TIME_DOW) {
        if (btn == BTN_UP) {
            if (messageCopy.dayOfWeek >= 7) messageCopy.dayOfWeek = 1;
            else messageCopy.dayOfWeek++;
        } else {
            if (messageCopy.dayOfWeek <= 1) messageCopy.dayOfWeek = 7;
            else messageCopy.dayOfWeek--;
        }
    } else if (timeCursor == CURSOR_TIME_DAY) {
        if (btn == BTN_UP) {
            if (messageCopy.day >= 31) messageCopy.day = 1;
            else messageCopy.day++;
        } else {
            if (messageCopy.day <= 1) messageCopy.day = 31;
            else messageCopy.day--;
        }
    } else if (timeCursor == CURSOR_TIME_MONTH) {
        if (btn == BTN_UP) {
            if (messageCopy.month >= 12) messageCopy.month = 1;
            else messageCopy.month++;
        } else {
            if (messageCopy.month <= 1) messageCopy.month = 12;
            else messageCopy.month--;
        }
    }
}

/*adjust alarm setup */
void adjustAlarmSetup(int btn) {
    int alarmNum = currentMode - MODE_ALARM_SETUP;

    if (alarmCursor == CURSOR_ALARM_ON_HOUR) {
        if (btn == BTN_UP) {
            if (alarms[alarmNum].onHour == 23) alarms[alarmNum].onHour = 255;
            else if (alarms[alarmNum].onHour == 255) alarms[alarmNum].onHour = 0;
            else alarms[alarmNum].onHour++;
        } else {
            if (alarms[alarmNum].onHour <= 0) alarms[alarmNum].onHour = 255;
            else if (alarms[alarmNum].onHour == 255) alarms[alarmNum].onHour = 23;
            else alarms[alarmNum].onHour--;
        }
    } else if (alarmCursor == CURSOR_ALARM_ON_MIN) {
        if (btn == BTN_UP) {
            if (alarms[alarmNum].onMinute == 59) alarms[alarmNum].onMinute = 255;
            else if (alarms[alarmNum].onMinute == 255) alarms[alarmNum].onMinute = 0;
            else alarms[alarmNum].onMinute++;
        } else {
            if (alarms[alarmNum].onMinute <= 0) alarms[alarmNum].onMinute = 255;
            else if (alarms[alarmNum].onMinute == 255) alarms[alarmNum].onMinute = 59;
            else alarms[alarmNum].onMinute--;
        }
    } else if (alarmCursor == CURSOR_ALARM_OFF_HOUR) {
        if (btn == BTN_UP) {
            if (alarms[alarmNum].offHour == 23) alarms[alarmNum].offHour = 255;
            else if (alarms[alarmNum].offHour == 255) alarms[alarmNum].offHour = 0;
            else alarms[alarmNum].offHour++;
        } else {
            if (alarms[alarmNum].offHour <= 0) alarms[alarmNum].offHour = 255;
            else if (alarms[alarmNum].offHour == 255) alarms[alarmNum].offHour = 23;
            else alarms[alarmNum].offHour--;
        }
    } else if (alarmCursor == CURSOR_ALARM_OFF_MIN) {
        if (btn == BTN_UP) {
            if (alarms[alarmNum].offMinute == 59) alarms[alarmNum].offMinute = 255;
            else if (alarms[alarmNum].offMinute == 255) alarms[alarmNum].offMinute = 0;
            else alarms[alarmNum].offMinute++;
        } else {
            if (alarms[alarmNum].offMinute <= 0) alarms[alarmNum].offMinute = 255;
            else if (alarms[alarmNum].offMinute == 255) alarms[alarmNum].offMinute = 59;
            else alarms[alarmNum].offMinute--;
        }
    } else if (alarmCursor == CURSOR_ALARM_ON_OFF_IND) {
        if (alarms[alarmNum].alarmSet == 0) alarms[alarmNum].alarmSet = 1;
        else alarms[alarmNum].alarmSet = 0;
    }
}

/* save settings and send the changed settings to the remote MCU via I2C */
void saveTimeSettings() {
    Wire.beginTransmission(ADDR_TWI_REMOTE);
    Wire.send(messageCopy.sender);
    Wire.send(messageCopy.msgType);
    Wire.send(messageCopy.year);
    Wire.send(messageCopy.month);
    Wire.send(messageCopy.day);
    Wire.send(messageCopy.hour);
    Wire.send(messageCopy.minute);
    Wire.send(messageCopy.secondInd);
    Wire.send(messageCopy.dayOfWeek);
    Wire.endTransmission();

    message = messageCopy;
}

/* save the alarm settings in EEPROM */
void saveAlarmSettings(int alarmNum) {
    int addr = alarmNum * 6;

    EEPROM.write(addr + 1, alarms[alarmNum].onHour);
    EEPROM.write(addr + 2, alarms[alarmNum].onMinute);
    EEPROM.write(addr + 3, alarms[alarmNum].offHour);
    EEPROM.write(addr + 4, alarms[alarmNum].offMinute);
    EEPROM.write(addr + 5, alarms[alarmNum].alarmSet);
}

/** <editor-fold desc="Arduino Methods"> */
void setup() {
    pinMode(PIN_ALARM, OUTPUT);
    pinMode(PIN_RELAY_1, OUTPUT);
    pinMode(PIN_RELAY_2, OUTPUT);
    pinMode(PIN_RELAY_3, OUTPUT);
    pinMode(PIN_RELAY_4, OUTPUT);
    pinMode(PIN_RELAY_5, OUTPUT);
    digitalWrite(PIN_ALARM, LOW);
    digitalWrite(PIN_RELAY_1, LOW);
    digitalWrite(PIN_RELAY_2, LOW);
    digitalWrite(PIN_RELAY_3, LOW);
    digitalWrite(PIN_RELAY_4, LOW);
    digitalWrite(PIN_RELAY_5, LOW);

    pinMode(PIN_BACKLIGHT, OUTPUT);

    for (int i = 0; i < NUM_OF_BUTTONS; i++) {
        pinMode(BTN_PINS[i], INPUT);
        digitalWrite(BTN_PINS[i], HIGH);
    }

    lcd.begin(16, 2);
    lcd.clear();
    lcd.print(defaultDisplayMsg);

    lcd.createChar(CLOCK_SYM, SYMBOL_CLOCK);
    lcd.createChar(CLOCK_SYM_SEL, SYMBOL_CLOCK_SEL);
    lcd.createChar(ALARM_SYM, SYMBOL_ALARM);
    lcd.createChar(ALARM_SYM_SEL, SYMBOL_ALARM_SEL);
    lcd.createChar(CTRL_SYM, SYMBOL_CTRL);
    lcd.createChar(CTRL_SYM_SEL, SYMBOL_CTRL_SEL);
    lcd.createChar(SAVED_SYM, SYMBOL_SAVED);

    lcd.setCursor(SYM_LINE_OFFSET, 1);
    lcd.write(CLOCK_SYM);

    for (int i = 0; i < NUM_OF_ALARMS; i++)
        lcd.write(ALARM_SYM);

    for (int i = 0; i < NUM_OF_RELAYS; i++)
        lcd.write(CTRL_SYM);

    Wire.begin(ADDR_TWI_LOCAL);
    Wire.onReceive(receiveEvent);

    getAlarms();
}

void loop() {
    //process button pressed events
    for (int i = 0; i < NUM_OF_BUTTONS; i++) {
        if (buttonPressed(i)) {
            if (intProc) return;
            cli();

            lcd.setCursor(15, 1);
            lcd.print(" ");

            if (i == BTN_MODE_SELECT) {
                currentMode++;
                renderCurrentMode();
            } else if (i == BTN_CURSOR_CTRL) {
                if (currentMode == MODE_TIME_SETUP) {
                    renderTimeSetup();
                } else if (currentMode >= MODE_ALARM_SETUP) {
                    renderAlarmSetup();
                }
            } else if (i == BTN_UP || i == BTN_DOWN) {
                if (currentMode == MODE_TIME_SETUP) {
                    adjustTimeSetup(i);
                } else if (currentMode >= MODE_ALARM_SETUP) {
                    adjustAlarmSetup(i);
                }
            } else if (i == BTN_ENTER) {
                if (currentMode == MODE_TIME_SETUP) {
                    sei();
                    saveTimeSettings();
                    cli();
                    lcd.setCursor(15, 1);
                    lcd.write(SAVED_SYM);
                } else if (currentMode >= MODE_ALARM_SETUP) {
                    saveAlarmSettings(currentMode - MODE_ALARM_SETUP);
                    lcd.setCursor(15, 1);
                    lcd.write(SAVED_SYM);
                } else if (currentMode == MODE_NORMAL_CANCEL) {
                    backLight = !backLight;

                    if (backLight) digitalWrite(PIN_BACKLIGHT, HIGH);
                    else digitalWrite(PIN_BACKLIGHT, LOW);
                }
            }
            sei();
        }


        if (buttonRleased(i)) {
            //the button is released, no action is necessary.
        }
    }

    checkAlarms();
}
/** </editor-fold> */